
import contextlib
import dataclasses
import logging
import os
import pkg_resources
import tempfile
import typing as t
import requests
from databind.core import annotations as A
from shut.model.requirements import Requirement
from shut.renderers.core import Renderer
from shut.test.base import BaseTestDriver, Runtime, TestRun, run_program_as_testcase
from shut.utils.io.virtual import VirtualFiles

if t.TYPE_CHECKING:
  from shut.model.package import PackageModel

log = logging.getLogger(__name__)


@dataclasses.dataclass
class RcfileSettings:

  DEFAULT_NAME = '.pylintrc'

  #: A template to use for the `.pylintrc` file. Must either be the name of a template delivered
  #: with shut (currently this is only `"shut"`) or a URL.
  template: t.Optional[str] = None

  #: If enabled, the rcfile will be rendered into the project directory on `shut pkg update`.
  render: bool = False

  #: The name under which to render the rcfile.
  name: t.Optional[str] = None

  # TODO (@NiklasRosenstein): Support overrides in the rcfile template.

  def __post_init__(self) -> None:
    if self.render and not self.template:
      raise ValueError(f'RcfileSettings.template must be set if RcfileSettings.render is enabled')

  def load_template(self) -> str:
    if not self.template:
      raise ValueError(f'RcfileSettings.template is not set')
    if self.template.startswith('http://') or self.template.startswith('https://'):
      response = requests.get(self.template)
      response.raise_for_status()
      return response.text
    try:
      return pkg_resources.resource_string('shut', f'data/pylintrc_templates/{self.template}.ini').decode('utf8')
    except FileNotFoundError:
      raise ValueError(f'.pylintrc template {self.template!r} does not exist')


@dataclasses.dataclass
class PylintTestDriver(BaseTestDriver, Renderer['PackageModel']):
  """
  Runs Pylint.
  """

  NAME = 'pylint'

  #: Environment variables when calling PyLint.
  env: t.Dict[str, str] = dataclasses.field(default_factory=dict)

  #: Additional arguments when calling Pylint.
  args: t.List[str] = dataclasses.field(default_factory=list)

  #: Options for the .pylintrc file to use when invoking Pylint.
  rcfile: t.Optional[RcfileSettings] = None

  def test_package(self, package: 'PackageModel', runtime: Runtime, capture: bool) -> TestRun:
    source_dir = package.get_source_directory()
    metadata = package.get_python_package_metadata()
    path = metadata.package_directory if not metadata.is_single_module else metadata.filename
    command = runtime.python + ['-m', 'pylint', path] + self.args

    with contextlib.ExitStack() as stack:

      if self.rcfile and (self.rcfile.name or self.rcfile.render):
        local_rcfile = os.path.join(package.get_directory(), self.rcfile.name or RcfileSettings.DEFAULT_NAME)
        command += ['--rcfile', local_rcfile]

      elif self.rcfile and self.rcfile.template:
        fp = stack.enter_context(tempfile.NamedTemporaryFile(delete=False, mode='w'))
        stack.callback(lambda: os.unlink(fp.name))
        fp.write(self.rcfile.load_template())
        fp.close()
        command += ['--rcfile', fp.name]

      test_run = run_program_as_testcase(
        runtime.get_environment(), source_dir, 'pylint',
        command=command, env=self.env, cwd=source_dir, capture=capture)

    return test_run

  def get_test_requirements(self) -> t.List[Requirement]:
    return [Requirement('pylint')]

  def get_files(self, files: VirtualFiles, obj: 'PackageModel') -> None:
    if self.rcfile and self.rcfile.render and self.rcfile.template:
      def render_rcfile(fp: t.TextIO) -> None:
        assert self.rcfile
        fp.write(self.rcfile.load_template())
      files.add_dynamic(self.rcfile.name or RcfileSettings.DEFAULT_NAME, render_rcfile)



# -*- coding: utf8 -*-
# Copyright (c) 2021 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Provides functions to render a list of changelogs. Always only supports the newest
changelog version.
"""

from shut.changelog.v2 import Entry
from .manager import Changelog
from nr.stream import Stream
from termcolor import colored
from typing import List, TextIO, Tuple
import re
import shutil
import textwrap


def _group_entries_by_component(entries: List[Entry]) -> List[Tuple[str, List[Entry]]]:
  return Stream(entries).sortby(lambda x: x.component).groupby(lambda x: x.component, lambda it: list(it)).collect()


def _terminal(fp: TextIO, changelogs: List[Changelog]) -> None:
  """
  Renders a #Changelog for the terminal to *fp*.
  """

  def _md_term_stylize(text: str) -> str:
    def _code(m):
      return colored(m.group(1), 'cyan')
    def _issue_ref(m):
      return colored(m.group(0), 'yellow', attrs=['bold'])
    text = re.sub(r'`([^`]+)`', _code, text)
    text = re.sub(r'#\d+', _issue_ref, text)
    return text

  def _fmt_issue(i):
    if str(i).isdigit():
      return '#' + str(i)
    return i

  def _fmt_issues(entry):
    if not entry.fixes:
      return None
    return '(' + ', '.join(colored(_fmt_issue(i), 'yellow', attrs=['underline']) for i in entry.fixes) + ')'

  def _fmt_types(entry):
    return colored(entry.type_.name, attrs=['bold'])

  if hasattr(shutil, 'get_terminal_size'):
    width = shutil.get_terminal_size((80, 23))[0]
  else:
    width = 80

  # Explode entries by component.
  for changelog in changelogs:
    fp.write(colored(str(changelog.version or 'Unreleased'), 'blue', attrs=['bold', 'underline']))
    fp.write(' ({})\n'.format(changelog.data.release_date or 'no release date'))
    for component, entries in _group_entries_by_component(changelog.entries):
      maxw = max(map(lambda x: len(x.type_.name), entries))
      fp.write('  ' + colored(component or 'No Component', 'red', attrs=['bold', 'underline']) + '\n')
      for entry in entries:
        lines = textwrap.wrap(entry.description, width - (maxw + 6))
        suffix_fmt = ' '.join(filter(bool, (_fmt_issues(entry),)))
        lines[-1] += ' ' + suffix_fmt
        delta = maxw - len(entry.type_.name)
        fp.write('    {} {}\n'.format(colored((_fmt_types(entry) + ':') + ' ' * delta, attrs=['bold']), _md_term_stylize(lines[0])))
        for line in lines[1:]:
          fp.write('    {}{}\n'.format(' ' * (maxw+2), _md_term_stylize(line)))
    fp.write('\n')


def _markdown(fp: TextIO, changelogs: List[Changelog]) -> None:

  def _fmt_issue(i):
    if str(i).isdigit():
      return '#' + str(i)
    return i

  def _fmt_issues(entry):
    if not entry.fixes:
      return None
    return '(' + ', '.join(_fmt_issue(i) for i in entry.fixes) + ')'

  for changelog in changelogs:
    fp.write('## {}'.format(changelog.version or 'Unreleased'))
    fp.write(' ({})\n\n'.format(changelog.data.release_date or 'no release date'))
    for component, entries in _group_entries_by_component(changelog.entries):
      fp.write('* __{}__\n'.format(component))
      for entry in entries:
        description ='**' + entry.type_.name + '**: ' + entry.description
        if entry.fixes:
          description += ' ' + _fmt_issues(entry)
        lines = textwrap.wrap(description, 80)
        fp.write('    * {}\n'.format(lines[0]))
        for line in lines[1:]:
          fp.write('      {}\n'.format(line))
    fp.write('\n')


renderers = {
  'terminal': _terminal,
  'markdown': _markdown,
}


def render(fp: TextIO, format: str, changelogs: List[Changelog]) -> None:
  changelogs = list(changelogs)
  unreleased = next((x for x in changelogs if not x.version), None)
  if unreleased:
    changelogs.remove(unreleased)
  changelogs.sort(key=lambda x: x.version, reverse=True)  # type: ignore
  if unreleased:
    changelogs.insert(0, unreleased)

  renderers[format](fp, changelogs)



# -*- coding: utf8 -*-
# Copyright (c) 2021 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import datetime
import json
import os
import subprocess as sp
import sys
from dataclasses import dataclass, field
from datetime import timezone
from typing import Dict, List, Optional, Tuple, TYPE_CHECKING
from typing_extensions import Annotated

from databind.core import annotations as A
from nr.parsing.date import tzlocal

from shut.model.requirements import Requirement
from .base import (BaseTestDriver, Runtime, StackTrace, TestCase, TestCrashReport,
  TestEnvironment, TestError, TestRun, TestStatus)

if TYPE_CHECKING:
  from shut.model.package import PackageModel


def load_report_file(report_file: str) -> TestRun:
  """
  Loads a Pytest-json-report file into a #TestRun.
  """

  with open(report_file) as fp:
    raw = json.load(fp)

  started_time = datetime.datetime.fromtimestamp(raw['created']).replace(tzinfo=tzlocal()).astimezone(timezone.utc)
  duration = raw['duration']
  status = TestStatus.PASSED if raw['exitcode'] == 0 else TestStatus.FAILED
  environment = TestEnvironment(raw['environment']['Python'], raw['environment']['Platform'])
  errors: List[TestError] = []
  tests: List[TestCase] = []

  # Map the nodes to the line and function that they are defined in.
  # NOTE (NiklasRosenstein): This key is not present when there are no test cases or
  #   when parallelizing tests with pytest-xdist.
  testid_to_source: Dict[str, Tuple[str, int]] = {}
  for node in raw.get('collectors', []):
    if node['nodeid'] and node['result']:
      for result in node['result']:
        if 'lineno' in result:
          testid_to_source[result['nodeid']] = (node['nodeid'], result['lineno'])
    if node['outcome'] != 'passed':
      errors.append(TestError(node['nodeid'], node['longrepr']))

  # Collect the test results.
  for test in raw['tests']:
    failed_stage = next((test[k] for k in ('setup', 'call', 'teardown')
      if k in test and test[k]['outcome'] == 'failed'), None)
    crash: Optional[TestCrashReport] = None
    stdout: str = ''
    if failed_stage:
      crash = TestCrashReport(
        filename=failed_stage['crash']['path'],
        lineno=failed_stage['crash']['lineno'],
        message=failed_stage['crash']['message'],
        traceback=[StackTrace(trace['path'], trace['lineno'], trace['message']) for trace in failed_stage['traceback']],
        longrepr=failed_stage['longrepr'],
      )
      stdout = failed_stage.get('stdout')
    test_status = {'passed': TestStatus.PASSED, 'failed': TestStatus.FAILED, 'skipped': TestStatus.SKIPPED}[test['outcome'].lstrip('x')]
    if test['nodeid'] in testid_to_source:
      filename, lineno = testid_to_source[test['nodeid']]
    else:
      filename = test['nodeid'].partition('::')[0]
      lineno = test['lineno']
    tests.append(TestCase(
      name=test['nodeid'],
      duration=sum(test[k]['duration'] for k in ('setup', 'call', 'teardown') if k in test),
      filename=filename,
      lineno=lineno,
      status=test_status,
      crash=crash,
      stdout=stdout,
    ))

  return TestRun(started_time, duration, status, environment, tests, errors)


@dataclass
class PytestDriver(BaseTestDriver):
  """
  A driver for running unit tests using [Pytest][1].

  [1]: https://docs.pytest.org/en/latest/
  """

  NAME = 'pytest'

  directory: Optional[str] = None
  args: List[str] = field(default_factory=lambda: ['-vv'])
  report_file: Annotated[str, A.alias('report-file')] = '.pytest-report.json'
  parallelism: int = 1

  # BaseTestDriver

  def test_package(self, package: 'PackageModel', runtime: Runtime, capture: bool) -> TestRun:
    test_dir = os.path.join(package.source_directory, self.directory) if self.directory else package.source_directory
    test_dir = os.path.join(package.get_directory(), test_dir)
    command = runtime.python + ['-m', 'pytest', test_dir]
    command += ['--json-report', '--json-report-file', self.report_file]
    if self.parallelism != 1:
      command += ['-n', str(self.parallelism)]
    command += self.args

    if os.path.isfile(self.report_file):
      os.remove(self.report_file)

    started_time = datetime.datetime.now(timezone.utc)
    proc = sp.Popen(command, stdout=sp.PIPE if capture else sys.stderr,
      stderr=sp.STDOUT if capture else None)
    output, _ = proc.communicate()
    result = proc.wait()

    if os.path.isfile(self.report_file):
      test_run = load_report_file(self.report_file)
      if result != 0:
        test_run.status = TestStatus.FAILED
      os.remove(self.report_file)
    else:
      test_run = TestRun(
        started=started_time,
        duration=(datetime.datetime.now(timezone.utc) - started_time).total_seconds(),
        status=TestStatus.ERROR,
        environment=runtime.get_environment(),
        tests=[],
        errors=[TestError(None, output.decode())],
      )

    return test_run

  def get_test_requirements(self) -> List[Requirement]:
    reqs = [
      Requirement.parse('pytest'),
      Requirement.parse('pytest-json-report ^1.2.1'),
    ]
    if self.parallelism > 1:
      reqs += [Requirement.parse('pytest-xdist')]
    return reqs



# -*- coding: utf8 -*-
# Copyright (c) 2021 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.


import datetime
import os
from dataclasses import dataclass, field
from typing import Dict, Iterable, Optional, Tuple, Union

import databind.json
import yaml

from shut.model import mapper
from shut.model.version import Version
from . import v1, v2, v3

AllChangelogTypes = Union[v3.Changelog, v2.Changelog, v1.Changelog]


@dataclass
class Changelog:
  """
  Represents a changelog on disk.
  """

  version: Optional[Version]
  data: v3.Changelog = field(default_factory=lambda: v3.Changelog(changes=[]))

  def __post_init__(self) -> None:
    self.filename: Optional[str] = None

  @property
  def entries(self):
    return self.data.changes

  def exists(self) -> bool:
    " Returns #True if the changelog file exists. "

    assert self.filename
    return os.path.isfile(self.filename)

  def load(self) -> None:
    " Loads the data from the file of this changelog. "

    assert self.filename
    with open(self.filename) as fp:
      raw_data = yaml.safe_load(fp)

    data: AllChangelogTypes = databind.json.load(raw_data, AllChangelogTypes, filename=self.filename, mapper=mapper)  # type: ignore
    if not isinstance(data, v3.Changelog):
      data = v3.Changelog.migrate(data)

    self.data = data

  def save(self, create_directory: bool = False) -> None:
    " Saves the changelog. It will always save the changelog in the newest supported format. "

    assert self.filename
    if create_directory:
      os.makedirs(os.path.dirname(self.filename), exist_ok=True)
    data = databind.json.dump(self.data, v3.Changelog, mapper=mapper)
    with open(self.filename, 'w') as fp:
      yaml.safe_dump(data, fp, sort_keys=False)

  def set_release_date(self, date: datetime.date) -> None:
    self.data.release_date = date

  def add_entry(self, entry) -> None:
    assert isinstance(entry, v3.Changelog.Entry), type(entry)
    self.data.changes.append(entry)


class ChangelogManager:

  def __init__(self, directory: str) -> None:
    self.directory = directory
    self._cache: Dict[Tuple[str, str], Changelog] = {}

  def _get(self, name: str, version: Optional[Version]) -> Changelog:
    key = (name, str(version))
    if key in self._cache:
      return self._cache[key]
    changelog = Changelog(version)
    changelog.filename = os.path.join(self.directory, name)
    if os.path.isfile(changelog.filename):
      changelog.load()
    self._cache[key] = changelog
    return changelog

  @property
  def unreleased(self) -> Changelog:
    return self._get('_unreleased.yml', None)

  def version(self, version: Version) -> Changelog:
    return self._get(str(version) + '.yml', version)

  def release(self, version: Version) -> Changelog:
    """
    Renames the unreleased changelog to the file name for the specified *version*.
    """

    unreleased = self.unreleased
    unreleased.data.release_date = datetime.date.today()
    unreleased.save()
    new_version = self.version(version)

    assert unreleased.filename
    assert new_version.filename
    os.rename(unreleased.filename, new_version.filename)
    self._cache.clear()

    return self.version(version)

  def all(self) -> Iterable[Changelog]:
    """
    Yields all changelogs.
    """

    for name in os.listdir(self.directory):
      if not name.endswith('.yml'):
        continue
      if name == '_unreleased.yml':
        yield self.unreleased
      else:
        version = Version(name[:-4])
        yield self.version(version)



# -*- coding: utf8 -*-
# Copyright (c) 2021 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from typing import Iterable, Generic, Type, TypeVar

T = TypeVar('T')


class TypeRegistry(Generic[T]):
  """
  A helper class to register objects with types. The registered objects can be retrieved
  back for a type, respecting registration and inheritance order.
  """

  def __init__(self):
    self._type_map = {}

  def put(self, type_: Type, data: Type[T]) -> None:
    self._type_map.setdefault(type_, []).append(data)

  def for_type(self, type_: Type) -> Iterable[Type[T]]:
    for base in type_.__bases__:
      yield from self.for_type(base)
    yield from self._type_map.get(type_, ())




import dataclasses
import datetime
import logging
import os
import subprocess as sp
import typing as t
from databind.core import annotations as A
from shut.model.requirements import Requirement
from shut.test.base import BaseTestDriver, Runtime, TestCase, TestCrashReport, TestEnvironment, TestRun, TestStatus, run_program_as_testcase

if t.TYPE_CHECKING:
  from shut.model.package import PackageModel

log = logging.getLogger(__name__)


@dataclasses.dataclass
class MypyTestDriver(BaseTestDriver):
  """
  Runs Mypy.
  """

  NAME = 'mypy'

  env: t.Dict[str, str] = dataclasses.field(default_factory=lambda: {'MYPY_FORCE_COLOR': '1'})
  args: t.List[str] = dataclasses.field(default_factory=lambda: ['--check-untyped-defs'])

  def test_package(self, package: 'PackageModel', runtime: Runtime, capture: bool) -> TestRun:
    source_dir = package.get_source_directory()
    command = runtime.python + ['-m', 'mypy']
    command += [source_dir] + self.args
    return run_program_as_testcase(
      runtime.get_environment(), source_dir, 'mypy',
      command=command, env=self.env, cwd=package.get_directory(), capture=capture)

  def get_test_requirements(self) -> t.List[Requirement]:
    return [Requirement('mypy')]



# -*- coding: utf8 -*-
# Copyright (c) 2021 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import abc
import enum
import datetime
import hashlib
import json
import logging
import os
import shlex
import shutil
import socket
import subprocess as sp
import sys
import textwrap
import traceback
import types
from dataclasses import dataclass, field
from typing import ClassVar, Dict, Optional, List, TYPE_CHECKING

from nr.preconditions import check_not_none
from databind.core import annotations as A
from shut.model.requirements import Requirement

if TYPE_CHECKING:
  from shut.model.package import PackageModel

log = logging.getLogger(__name__)

__all__ = [
  'TestStatus',
  'TestEnvironment',
  'StackTrace',
  'TestCrashReport',
  'TestCase',
  'TestRun',
  'Runtime',
  'Virtualenv',
  'BaseTestDriver',
]


class TestStatus(enum.Enum):
  PASSED = enum.auto()  #: The test passed.
  FAILED = enum.auto()  #: The test failed.
  SKIPPED = enum.auto()  #: The test was skipped.
  ERROR = enum.auto()  #: An error occurred when running the test.


@dataclass
class TestEnvironment:
  python_version: str
  platform: Optional[str]


@dataclass
class StackTrace:
  filename: str
  lineno: int
  message: str

  @classmethod
  def from_traceback(cls, tb: Optional[types.TracebackType]) -> List['StackTrace']:
    result = []
    while tb:
      result.append(cls(filename=tb.tb_frame.f_code.co_filename, lineno=tb.tb_frame.f_lineno, message=''))
      tb = tb.tb_next
    return result


@dataclass
class TestCrashReport:
  filename: str
  lineno: int
  message: str
  traceback: List[StackTrace]
  longrepr: str

  @classmethod
  def current_exception(cls) -> 'TestCrashReport':
    exc = sys.exc_info()
    if exc is None:
      raise RuntimeError('no current exception')
    tb = exc[2]
    assert tb
    return cls(
      filename=tb.tb_frame.f_code.co_filename,
      lineno=tb.tb_frame.f_lineno,
      message=str(exc[1]),
      traceback=StackTrace.from_traceback(tb),
      longrepr='\n'.join(traceback.format_exception(*exc)))


@dataclass
class TestCase:
  name: str
  duration: float
  filename: str
  lineno: int
  status: TestStatus
  crash: Optional[TestCrashReport]
  stdout: str


@dataclass
class TestError:
  filename: Optional[str]
  longrepr: str


@dataclass
class TestRun:
  started: datetime.datetime
  duration: float
  status: TestStatus
  environment: TestEnvironment
  tests: List[TestCase]
  errors: List[TestError] = field(default_factory=list)
  error: Optional[str] = None


@dataclass
class Runtime:
  python: List[str]
  pip: List[str]
  virtualenv: List[str]

  def __post_init__(self) -> None:
    self._env_info: Optional[Dict[str, Optional[str]]] = None

  @classmethod
  def from_env(self) -> 'Runtime':
    python = shlex.split(os.getenv('PYTHON', 'python'))
    pip_var = os.getenv('PIP')
    pip = shlex.split(pip_var) if pip_var else python + ['-m', 'pip']
    venv_var = os.getenv('VIRTUALENV')
    virtualenv = shlex.split(venv_var) if venv_var else python + ['-m', 'venv']
    return Runtime(python, pip, virtualenv)

  @classmethod
  def from_python3(self, python: List[str]) -> 'Runtime':
    pip = python + ['-m', 'pip']
    virtualenv = python + ['-m', 'venv']
    return Runtime(python, pip, virtualenv)

  def is_venv(self) -> bool:
    """
    Returns #True if the #Runtime refers to a local environment (e.g. Virtualenv).

    Todo: Detect Conda environments as local environments?
    """

    # See https://stackoverflow.com/a/42580137/791713
    info = self._get_environment_info()
    return bool(info['real_prefix'] or (info['base_prefix'] and info['prefix'] != info['base_prefix']))

  def get_executable_path(self) -> str:
    return check_not_none(self._get_environment_info()['executable'])

  def get_hash_code(self) -> str:
    """
    Returns a has that identifies the runtime. This is a combination of the environment info
    and the creation timestamp of the Python executable (which we expect to only be changed if
    a virtualenv is re-created).
    """

    path = self.get_executable_path()
    ctime = os.path.getctime(path)
    return hashlib.md5(f'{socket.gethostname()}-{path}-{ctime}'.encode()).hexdigest()

  def _get_environment_info(self) -> Dict[str, Optional[str]]:
    if self._env_info is None:
      code = textwrap.dedent('''
        import sys, platform, json
        print(json.dumps({
          "executable": sys.executable,
          "version": sys.version,
          "platform": platform.platform(),
          "prefix": sys.prefix,
          "base_prefix": getattr(sys, 'base_prefix', None),
          "real_prefix": getattr(sys, 'real_prefix', None),
        }))
      ''')
      self._env_info = json.loads(sp.check_output(self.python + ['-c', code]).decode())
    return self._env_info

  def get_environment(self) -> TestEnvironment:
    raw = self._get_environment_info()
    version, platform = raw['version'], raw['platform']
    assert version and platform
    return TestEnvironment(version, platform)


@dataclass
class Virtualenv:
  path: str

  def exists(self) -> bool:
    return os.path.isdir(self.path)

  def create(self, runtime: Runtime) -> None:
    """
    Create the virtual environment.
    """

    sp.check_call(runtime.virtualenv + [self.path])

  def rm(self):
    if os.path.isdir(self.path):
      shutil.rmtree(self.path)

  def bin(self, name):
    if os.name == 'nt':
      name += '.exe'
    return os.path.join(self.path, 'Scripts' if os.name == 'nt' else 'bin', name)

  def get_runtime(self) -> Runtime:
    return Runtime.from_python3([self.bin('python')])


@A.union(A.union.Subtypes.entrypoint('shut.test_drivers'), style=A.union.Style.flat)
class BaseTestDriver(abc.ABC):
  """
  Base class for drivers that can run unit tests for a package.
  """

  NAME: ClassVar[str]

  @abc.abstractmethod
  def test_package(self, package: 'PackageModel', runtime: Runtime, capture: bool) -> TestRun:
    pass

  @abc.abstractmethod
  def get_test_requirements(self) -> List[Requirement]:
    pass


def run_program_as_testcase(
  environment: TestEnvironment,
  filename: str,
  test_run_name: str,
  command: List[str],
  env: Dict[str, str],
  cwd: Optional[str] = None,
  capture: bool = True,
) -> TestRun:
  """
  Run a program as a #TestRun.
  """

  proc_env = os.environ.copy()
  proc_env.update(env)

  log.debug('Running command %s in %s', command, cwd or os.getcwd())

  crash: Optional[TestCrashReport] = None
  started = datetime.datetime.now()

  if capture:
    stdout, stderr, stdin = sp.PIPE, sp.STDOUT, sp.DEVNULL
  else:
    stdout, stderr, stdin = 1, 2, sp.DEVNULL

  try:
    proc = sp.Popen(command, stdout=stdout, stderr=stderr, stdin=stdin, env=proc_env, cwd=cwd)
    output = (proc.communicate()[0] or b'').decode()
  except OSError:
    status = TestStatus.ERROR
    crash = TestCrashReport.current_exception()
    output = ''
  else:
    status = TestStatus.PASSED if proc.returncode == 0 else TestStatus.FAILED

  duration = (datetime.datetime.now() - started).total_seconds()

  return TestRun(
    started=started,
    duration=duration,
    status=status,
    environment=environment,
    tests=[TestCase(
      name=test_run_name,
      duration=duration,
      filename=filename,
      lineno=0,
      status=status,
      crash=crash,
      stdout=output,
    )]
  )



# -*- coding: utf8 -*-
# Copyright (c) 2021 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
import os
from typing import Optional

import click

from shut.commands import project
from shut.commands.commons.new import write_files
from shut.commands.pkg.update import verify_integrity, verify_tag, _VERIFY_TAG_HELP
from shut.model import MonorepoModel
from shut.renderers import get_files
from shut.utils.io.virtual import VirtualFiles
from . import mono


def update_monorepo(
  monorepo: MonorepoModel,
  dry: bool = False,
  indent: int = 0,
  verify: bool = False,
  tag: Optional[str] = None,
  all_: bool = False,
) -> VirtualFiles:

  result = 0
  if tag is not None and not verify_tag(monorepo, tag):
    result = 1

  assert monorepo.project
  files = get_files(monorepo)
  if all_:
    for package in monorepo.project.packages:
      files.update(get_files(package), os.path.basename(package.get_directory()))

  if verify and not verify_integrity(monorepo, files):
    result = 1
  if result != 0:
    sys.exit(result)
  if not verify:
    write_files(files, monorepo.get_directory(), force=True, dry=dry, indent=indent)
  return files


@mono.command()
@click.option('--dry', is_flag=True)
@click.option('--verify', is_flag=True, help='Verify the integrity of the managed files '
  '(asserting that they would not change from running this command).')
@click.option('--verify-tag', help=_VERIFY_TAG_HELP)
def update(dry: bool, verify: bool, verify_tag: Optional[str]) -> None:
  """
  Update files auto-generated from the configuration file.
  """

  if verify_tag is not None and verify_tag.startswith('refs/tags/'):
    verify_tag = verify_tag[10:]

  monorepo = project.load_or_exit(expect=MonorepoModel)
  update_monorepo(monorepo, dry, verify=verify or verify_tag is not None, tag=verify_tag, all_=True)



# -*- coding: utf8 -*-
# Copyright (c) 2021 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import ast
import os
import re
import shlex
import warnings
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, List, Optional, Union, cast
from typing_extensions import Annotated

from databind.core import annotations as A
from nr.stream import Stream

from shut.utils.ast import load_module_members
from shut.utils.fs import get_file_in_directory
from .author import Author
from .abstract import AbstractProjectModel
from .linter import LinterConfiguration
from .publish import PublishConfiguration
from .requirements import Requirement, RequirementsList
from ..test import BaseTestDriver
from .version import Version

if TYPE_CHECKING:
  from shut.renderers.core import Renderer


class PackageError(Exception):
  pass


@dataclass
class InstallConfiguration:
  @dataclass
  class InstallHooks:
    before_install: Annotated[List[str], A.alias('before-install')] = field(default_factory=list)
    after_install: Annotated[List[str], A.alias('after-install')] = field(default_factory=list)
    before_develop: Annotated[List[str], A.alias('before-develop')] = field(default_factory=list)
    after_develop: Annotated[List[str], A.alias('after-develop')] = field(default_factory=list)

    def any(self):
      return any((self.before_install, self.after_install, self.before_develop, self.after_develop))

    def as_payload(self) -> Dict[str, List[str]]:
      def _splitall(x):
        return list(map(shlex.split, x))
      return {
        'before-install': _splitall(self.before_install),
        'after-install': _splitall(self.after_install),
        'before-develop': _splitall(self.before_develop),
        'after-develop': _splitall(self.after_develop),
      }

  hooks: InstallHooks = field(default_factory=InstallHooks)

  #: A value for the `--index-url` option to pass to Pip when using `shut pkg install`.
  index_url: Annotated[Optional[str], A.alias('index-url')] = field(default=None)

  #: A list of URLs to pass to Pip via the `--extra-index-url` option when using
  #: `shut pkg install`.
  extra_index_urls: Annotated[List[str], A.alias('extra-index-urls')] = field(default_factory=list)

  def get_pip_args(self) -> List[str]:
    result = []
    if self.index_url:
      result += ['--index-url', self.index_url]
    for url in self.extra_index_urls:
      result += ['--extra-index-url', url]
    return result


@dataclass
class Include:
  include: str


@dataclass
class Exclude:
  exclude: str


@dataclass
class PackageModel(AbstractProjectModel):
  modulename: Optional[str] = None
  description: Optional[str] = None
  readme: Optional[str] = None
  wheel: Optional[bool] = True
  universal: Optional[bool] = None
  typed: Optional[bool] = False
  requirements: RequirementsList = field(default_factory=RequirementsList)
  test_requirements: Annotated[RequirementsList, A.alias('test-requirements')] = field(default_factory=RequirementsList)
  extra_requirements: Annotated[Dict[str, RequirementsList], A.alias('extra-requirements')] = field(default_factory=dict)
  dev_requirements: Annotated[RequirementsList, A.alias('dev-requirements')] = field(default_factory=RequirementsList)
  render_requirements_txt: Annotated[bool, A.alias('render-requirements-txt')] = False
  source_directory: Annotated[str, A.alias('source-directory')] = 'src'
  exclude: List[str] = field(default_factory=lambda: ['test', 'tests', 'docs'])
  entrypoints: Dict[str, List[str]] = field(default_factory=dict)
  classifiers: List[str] = field(default_factory=list)
  keywords: List[str] = field(default_factory=list)
  package_data: Annotated[List[Union[Include, Exclude]], A.alias('package-data')] = field(default_factory=list)

  install: InstallConfiguration = field(default_factory=InstallConfiguration)
  linter: LinterConfiguration = field(default_factory=LinterConfiguration)
  publish: PublishConfiguration = field(default_factory=PublishConfiguration)

  #: Deprecated, use #test_drivers instead.
  test_driver: Annotated[Optional[BaseTestDriver], A.alias('test-driver')] = None
  test_drivers: Annotated[List[BaseTestDriver], A.alias('test-drivers')] = field(default_factory=list)

  def validate(self) -> None:
    if self.test_driver:
      warnings.warn(f'({self._filename}) $.test-driver is deprecated since version 0.17.0, please use $test-drivers instead.')

  def get_modulename(self) -> str:
    if self.modulename:
      return self.modulename

    # Check for PEP-561 stub packages.
    # See also https://mypy.readthedocs.io/en/latest/installed_packages.html#making-pep-561-compatible-packages
    if self.name.endswith('-stubs'):
      return self.name[:-6].replace('-', '_') + '-stubs'

    return self.name.replace('-', '_')

  def get_python_requirement(self) -> Optional[Requirement]:
    return next(filter(lambda x: isinstance(x, Requirement) and x.package == 'python', self.requirements), None)  # type: ignore

  def has_vendored_requirements(self) -> bool:
    """
    Returns #True if the package has any vendored requirements.
    """

    return any(Stream([
      self.requirements.vendored_reqs(),
      self.test_requirements.vendored_reqs(),
      *(l.vendored_reqs() for l in self.extra_requirements.values())
    ]).concat())

  def is_universal(self) -> bool:
    """
    Checks if the package is a universal Python package (i.e. it is Python 2 and 3 compatible)
    by testing the `$.requirements.python` version selector. If none is specified, the package
    is also considered universal.
    """

    if self.universal is not None:
      return self.universal

    python_requirement = self.get_python_requirement()
    if not python_requirement:
      return True

    # TODO (@NiklasRosenstein): This method of detecting if the version selector
    #   selects a Python 2 and 3 version is very suboptimal.
    has_2 = re.search(r'\b2\b|\b2\.\b', str(python_requirement))
    has_3 = re.search(r'\b3\b|\b3\.\b', str(python_requirement))
    return bool(has_2 and has_3)

  def get_python_package_metadata(self) -> 'PythonPackageMetadata':
    """
    Returns a #PythonPackageMetadata object for this #PackageModel. This object can be
    used to inspect the author and version information that is defined in the package
    source code.
    """

    assert self.filename
    return PythonPackageMetadata(
      os.path.join(os.path.dirname(self.filename), self.source_directory),
      self.get_modulename())

  def get_readme_file(self) -> Optional[str]:
    """
    Returns the absolute path to the README for this package.
    """

    assert self.filename
    directory = os.path.dirname(self.filename)

    if self.readme:
      return os.path.abspath(os.path.join(directory, self.readme))

    return get_file_in_directory(
      directory=directory,
      prefix='README.',
      preferred=['README.md', 'README.rst', 'README.txt', 'README'],
      case_sensitive=False)

  def get_py_typed_file(self) -> Optional[str]:
    if not self.typed:
      return None

    directory = self.get_python_package_metadata().package_directory
    return os.path.join(directory, 'py.typed')

  def get_publish_config(self) -> PublishConfiguration:
    assert self.project
    if self.project and self.project.monorepo and self.project.monorepo.publish:
      return self.project.monorepo.publish
    return self.publish

  def get_license(self) -> Optional[str]:
    assert self.project
    if self.license:
      return self.license
    if self.project.monorepo:
      return self.project.monorepo.license
    return None

  def get_author(self) -> Optional[Author]:
    assert self.project
    if self.author:
      return self.author
    if self.project.monorepo:
      return self.project.monorepo.author
    return None

  def get_url(self) -> Optional[str]:
    assert self.project
    if self.url:
      return self.url
    if self.project.monorepo:
      return self.project.monorepo.url
    return None

  def get_source_directory(self) -> str:
    return os.path.join(self.get_directory(), self.source_directory)

  def get_test_drivers(self) -> List[BaseTestDriver]:
    result = []
    if self.test_driver:
      result.append(self.test_driver)
    result += self.test_drivers
    return result

  def get_auxiliary_renderers(self) -> List['Renderer[PackageModel]']:
    from shut.renderers.core import Renderer
    result = []
    for driver in self.get_test_drivers():
      if isinstance(driver, Renderer):
        result.append(cast('Renderer[PackageModel]', driver))
    return result

  # AbstractProjectModel

  def get_name(self) -> str:
    return self.name

  def get_version(self) -> Optional[Version]:
    assert self.project
    if self.version:
      return self.version
    if self.project.monorepo and self.project.monorepo.release.single_version:
      return self.project.monorepo.get_version()
    return None

  def get_tag(self, version: Version) -> str:
    assert self.project
    tag_format = self.release.tag_format
    if self.project and self.project.monorepo and '{name}' not in tag_format:
      tag_format = '{name}@' + tag_format
    return tag_format.format(name=self.name, version=version)

  def get_license_file(self, inherit: bool = False) -> Optional[str]:
    assert self.project
    if not self.license_file and inherit and self.project.monorepo and \
        (not self.license or self.license == self.project.monorepo.license):
      filename = self.project.monorepo.get_license_file()
      if filename:
        return os.path.join(self.project.monorepo.get_directory(), filename)
    return super().get_license_file(False)


class PythonPackageMetadata:
  """
  Represents the metadata of a Python package on disk.
  """

  def __init__(self, source_directory: str, modulename: str) -> None:
    self.source_directory = source_directory
    self.modulename = modulename
    self._filename: Optional[str] = None
    self._author: Optional[str] = None
    self._version: Optional[str] = None

  @property
  def filename(self) -> str:
    """
    Returns the file that contains the package metadata in the Python source code. This is
    usually the module filename, the package `__init__.py` or `__version__.py`.
    """

    if self._filename:
      return self._filename

    parts = self.modulename.split('.')
    prefix = os.sep.join(parts[:-1])
    choices = [
      parts[-1] + '.py',
      os.path.join(parts[-1], '__version__.py'),
      os.path.join(parts[-1], '__init__.py'),
    ]
    for filename in choices:
      filename = os.path.join(self.source_directory, prefix, filename)
      if os.path.isfile(filename):
        self._filename = filename
        return filename

    raise PackageError('Entry file for package "{}" could not be determined'
                       .format(self.modulename))

  @property
  def package_directory(self) -> str:
    """
    Returns the Python package directory. Raises a #ValueError if this metadata represents
    just a single Python module.
    """

    if self.is_single_module:
      raise ValueError('this package is in module-only form')

    return os.path.dirname(self.filename) or '.'

  @property
  def is_single_module(self) -> bool:
    basename = os.path.basename(self.filename)
    return basename not in ('__init__.py', '__version__.py')

  @property
  def author(self) -> str:
    if not self._author:
      self._load_metadata()
    assert self._author
    return self._author

  @property
  def version(self) -> str:
    if not self._version:
      self._load_metadata()
    assert self._version
    return self._version

  def _load_metadata(self) -> None:
    members = load_module_members(self.filename)

    author = None
    version = None

    if '__version__' in members:
      try:
        version = ast.literal_eval(members['__version__'])
      except ValueError:
        version = '<Non-literal expression>'

    if '__author__' in members:
      try:
        author = ast.literal_eval(members['__author__'])
      except ValueError:
        author = '<Non-literal expression>'

    self._author = author
    self._version = version




import typing as t
from databind.core import Converter, Context, ConcreteType, Direction
from databind.core.types.schema import ObjectType, dataclass_to_schema
from databind.json.modules.object import ObjectTypeConverter


class StringConverter(Converter):

  def __init__(self, loader: t.Optional[t.Callable[[str], t.Any]] = None) -> None:
    self._loader = loader
    self._fallback = ObjectTypeConverter()

  def convert(self, ctx: Context) -> object:
    assert isinstance(ctx.type, ConcreteType)
    if ctx.direction == Direction.serialize:
      assert isinstance(ctx.value, ctx.type.type)
      return str(ctx.value)
    else:
      if isinstance(ctx.value, t.Mapping):
        return self._fallback.convert(ctx.push(ObjectType(
          dataclass_to_schema(ctx.type.type, ctx.type_hint_adapter), []), ctx.value, None))
      elif not isinstance(ctx.value, str):
        raise ctx.type_error(expected=str)
      if self._loader:
        return self._loader(ctx.value)
      else:
        return ctx.type.type.from_string(ctx.value)



# -*- coding: utf8 -*-
# Copyright (c) 2021 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import abc
import enum
import os
import posixpath
import re
import typing as t
from dataclasses import dataclass
from typing import Iterable, List, Optional, Union, TypeVar, Type
from urllib.parse import urlparse

from .version import bump_version, Version

T = TypeVar('T')


class VersionSelector(object):
  """
  Represents a version selector that limits which versions of a requirement can be
  picked. It simply wraps a string, but if the string looks like a semver selector,
  it can be converted into a ``setuptools`` compatible version selector string (see
  #to_setuptools()).
  """

  ANY: 'VersionSelector'

  def __init__(self, selector: t.Union[str, 'VersionSelector']):
    if isinstance(selector, VersionSelector):
      selector = selector._string
    self._string: str = selector.strip()

  def __str__(self):
    return str(self._string)

  def __repr__(self):
    return 'VersionSelector({!r})'.format(self._string)

  def __eq__(self, other):
    if type(self) == type(other):
      return self._string == other._string
    return False

  def __ne__(self, other):
    return not (self == other)

  def to_setuptools(self):  # type: () -> str
    """
    Converts the version selector to a string that Setuptools/Pip can understand by
    expanding the `~` and `^` range selectors.

    Given a version number X.Y.Z, the selectors will be expanded as follows:

    - `^X.Y.Z` -> `>=X.Y.Z,<X+1.0.0`
    - `~X.Y.Z` -> `>=X.Y.Z,<X.Y+1.0`
    - `X.Y.Z -> ==X.Y.Z`
    """

    # Poor-mans test if this looks like the form 'X.Y.Z' without anything around it.
    if not ',' in self._string and self._string[0].isdigit():
      return '==' + self._string

    regex = r'[~^](\d+\.\d+(\.\d+)?[.\-\w]*)'
    def sub(match):
      index = {'^': 0, '~': 1}[match.group(0)[0]]
      max_version = match.group(1).split('.')[:3]
      if len(max_version) == 2:
        max_version.append('0')
      if '-' in max_version[-1]:
        max_version[-1] = max_version[-1].partition('-')[0]
      max_version[index] = str(int(max_version[index]) + 1)
      for i in range(index+1, 3):
        max_version[i] = '0'
      return '>={},<{}'.format(match.group(1), '.'.join(max_version))

    s = self._string + '.0' * (3 - self._string.count('.') - 1)
    return re.sub(regex, sub, s)

  def is_semver_selector(self) -> bool:
    return bool(self._string and self._string[0] in '^~' and ',' not in self._string)

  def matches(self, version: Union[Version, str]) -> bool:
    if not self.is_semver_selector():
      # TODO (@NiklasRosenstein): Match setuptools version selectors.
      return False
    min_version = Version(self._string[1:])
    if self._string[0] == '^':
      max_version = bump_version(min_version, 'major')
    elif self._string[0] == '~':
      max_version = bump_version(min_version, 'minor')
    else:
      raise RuntimeError('invalid semver selector string {!r}'.format(self._string))
    return min_version <= Version(version) < max_version


VersionSelector.ANY = VersionSelector('*')


class BaseRequirement(abc.ABC):

  @abc.abstractclassmethod
  def from_string(cls, value: str) -> 'BaseRequirement': ...

  @abc.abstractmethod
  def __str__(self) -> str: ...

  @abc.abstractmethod
  def to_setuptools(self) -> str: ...


@dataclass
class Requirement(BaseRequirement):
  """
  A Requirement is simply combination of a package name and a version selector.
  """

  INVALID_CHARACTERS = '/&:;\'"!{}()%'

  package: str
  version: VersionSelector = VersionSelector.ANY
  extras: Optional[List[str]] = None

  def __str__(self, *, setuptools: bool = False):
    result = self.package
    if self.extras:
      result += '[' + ','.join(sorted(self.extras)) + ']'
    if self.version != VersionSelector.ANY:
      result += ' ' + (str(self.version.to_setuptools()) if setuptools else str(self.version))
    return result

  def __repr__(self):
    return repr(str(self))

  @classmethod
  def parse(cls, requirement_string: str) -> 'Requirement':
    error = ValueError('invalid requirement: {!r}'.format(requirement_string))
    if set(requirement_string) & set(cls.INVALID_CHARACTERS):
      raise error
    match = re.match(r'''
      ^\s*                # Allow leading whitespace
      ([\w\d\-\._]+)      # Package name
      (?:\[([^\]]+)\])?   # Package extras
      (?:\s*([^\[\]]+))?  # Version selector (do not allow brackets to avoid matching empty extras)
      $''', requirement_string, re.VERBOSE)
    if not match:
      raise error
    package, extras, version = match.groups()
    return cls(package, VersionSelector(version or VersionSelector.ANY),
      extras=sorted(map(str.strip, extras.split(','))) if extras else None)

  def to_setuptools(self) -> str:
    return self.__str__(setuptools=True)  # type: ignore

  @classmethod
  def from_string(cls, value: str) -> 'Requirement':
    return Requirement.parse(value)


@dataclass
class VendoredRequirement(BaseRequirement):
  """
  A vendored requirement is either a relative path or a string prefixed with `git+` that
  Pip understands as an installable source.
  """

  INVALID_CHARACTERS = '*&^;\'"!{}[]()%'

  class Type(enum.Enum):
    Git = enum.auto()
    Path = enum.auto()

  type: Type
  location: str

  def __post_init__(self):
    if self.type == self.Type.Path:
      self.location = self._normpath(self.location)

  def __str__(self):
    if self.type == self.Type.Git:
      return 'git+' + self.location
    elif self.type == self.Type.Path:
      return self._normpath(self.location)
    assert False, self.type

  @staticmethod
  def _normpath(path: str) -> str:
    """
    Normalizes a path in the way that it should be represented in the serialized form of the
    vendored requirement (as posix path).
    """

    if os.name == 'nt':
      path = path.replace('\\', '/')
    if not os.path.isabs(path):
      path = './' + posixpath.normpath(path)
    return path

  @classmethod
  def parse(cls, requirement_string: str, fallback_to_path: bool = False) -> 'VendoredRequirement':
    error = ValueError(f'invalid vendored requirement: {requirement_string!r}')
    if set(requirement_string) & set(cls.INVALID_CHARACTERS):
      raise error
    if requirement_string.startswith('git+'):
      url = requirement_string[4:]
      if not urlparse(url).scheme:
        raise error
      return cls(cls.Type.Git, url)
    result = urlparse(requirement_string)
    if result.scheme:
      raise error
    if not requirement_string.startswith('./') and not fallback_to_path:
      raise error
    return cls(cls.Type.Path, requirement_string)

  def to_setuptools(self) -> str:
    raise RuntimeError('VendoredRequirement is not supported in setuptools')

  def get_pip_args(self, root: str, develop: bool) -> List[str]:
    args = [self.location]
    if self.type == self.Type.Path:
      args[0] = os.path.normpath(os.path.join(root, args[0]))
      # We gotta make sure that Pip recognizes the argument as a path if there's no path
      # separator in there.
      if args[0].count(os.sep) == 0:
        args[0] = os.path.join(os.curdir, args[0])
      if develop:
        args.insert(0, '-e')
    return args

  @classmethod
  def from_string(cls, value: str) -> 'VendoredRequirement':
    return VendoredRequirement.parse(value)


class RequirementsList(List[Union[Requirement, VendoredRequirement]]):
  """
  A list of requirements. Provides some convenience methods.
  """

  def reqs(self) -> Iterable[Requirement]:
    return filter(lambda x: isinstance(x, Requirement), self)  # type: ignore

  def vendored_reqs(self) -> Iterable[VendoredRequirement]:
    return filter(lambda x: isinstance(x, VendoredRequirement), self)  # type: ignore

  def get_pip_args(self, root: str, develop: bool) -> List[str]:
    result = []
    for req in self.reqs():
      result.append(req.to_setuptools())
    for vreq in self.vendored_reqs():
      result += vreq.get_pip_args(root, develop)
    return result


from .utils import StringConverter
from . import mapper
mapper.add_converter_for_type(Requirement, StringConverter())  # type: ignore
mapper.add_converter_for_type(VendoredRequirement, StringConverter())  # type: ignore



# -*- coding: utf8 -*-
# Copyright (c) 2021 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from dataclasses import dataclass, field
import abc
import os
import warnings
from typing import TYPE_CHECKING, List, Optional, TypeVar

from .author import Author
from .changelog import ChangelogConfiguration
from .release import ReleaseConfiguration
from .version import Version
from shut.utils.fs import get_file_in_directory

if TYPE_CHECKING:
  from shut.renderers.core import Renderer

T_AbstractProjectModel = TypeVar('T_AbstractProjectModel', bound='AbstractProjectModel')


@dataclass
class AbstractProjectModel(abc.ABC):
  name: str
  version: Optional[Version] = None
  author: Optional[Author] = None
  license: Optional[str] = None
  license_file: Optional[str] = None
  url: Optional[str] = None
  changelog: ChangelogConfiguration = field(default_factory=ChangelogConfiguration)
  release: ReleaseConfiguration = field(default_factory=ReleaseConfiguration)

  def __post_init__(self) -> None:
    # May be filled during the deserialization process to track additional metadata.
    self._project: Optional['Project'] = None
    self._filename: Optional[str] = None
    self.unknown_keys: List[str] = []

  def validate(self) -> None: ...

  @property
  def project(self) -> 'Project':
    if self._project is None:
      raise RuntimeError(f'{type(self).__name__}.project is not set')
    return self._project

  @project.setter
  def project(self, project: 'Project') -> None:
    self._project = project

  @property
  def filename(self) -> str:
    if self._filename is None:
      raise RuntimeError(f'{type(self).__name__}.filename is not set')
    return self._filename

  @filename.setter
  def filename(self, filename: str) -> None:
    self._filename = filename

  def get_name(self) -> str:
    warnings.warn(f'{type(self).__name__}.get_name() is deprecated, use .name instead', DeprecationWarning)
    return self.name

  def get_version(self) -> Optional[Version]:
    warnings.warn(f'{type(self).__name__}.get_version() is deprecated, use .version instead', DeprecationWarning)
    return self.version

  def get_tag(self, version: Version) -> str:
    return self.release.tag_format.format(name=self.get_name(), version=version)

  def get_directory(self) -> str:
    assert self.filename
    return os.path.dirname(self.filename)

  def get_author(self) -> Optional[Author]:
    return self.author

  def get_changelog_directory(self) -> str:
    assert self.filename
    return os.path.join(os.path.dirname(self.filename), self.changelog.directory)

  def get_license_file(self, inherit: bool = False) -> Optional[str]:
    """
    Returns the absolute path to the LICENSE file for this package.
    """

    assert self.filename
    if self.license_file:
      return self.license_file

    return get_file_in_directory(
      directory=os.path.dirname(self.filename),
      prefix='LICENSE.',
      preferred=['LICENSE', 'LICENSE.txt', 'LICENSE.rst', 'LICENSE.md'],
      case_sensitive=False)

  def get_auxiliary_renderers(self: T_AbstractProjectModel) -> List['Renderer[T_AbstractProjectModel]']:
    return []

from . import Project  # pylint: disable=unused-import



# -*- coding: utf8 -*-
# Copyright (c) 2021 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import logging
import sys
from typing import List, Optional

import click
import yaml
import databind.json
from nr.utils.git import Git
from termcolor import colored

from .. import shut, commons, project
from shut.changelog import v3
from shut.changelog.manager import ChangelogManager
from shut.changelog.render import render as render_changelogs
from shut.model import mapper
from shut.model.version import parse_version
from shut.utils.cli import editor_open, edit_text
from shut.utils.functional import expect


_git = Git()
logger = logging.getLogger(__name__)


@shut.command()
@click.argument('version', type=parse_version, required=False)
@click.option('--reformat', is_flag=True, help='reformat the changelog')
@click.option('--add', metavar='type', help='create a new changelog entry')
@click.option('--for', metavar='component', help='components for the new changelog entry (default: general)', default='general')
@click.option('--fixes', metavar='issue,…', help='issues that this changelog entry fixes')
@click.option('-m', '--message', metavar='text', help='changelog entry description')
@click.option('-e', '--edit', is_flag=True, help='edit the changelog entry or file')
@click.option('--markdown', is_flag=True, help='render the changelog as markdown')
@click.option('-a', '--all', is_flag=True, help='show the changelog for all versions')
@click.option('-s', '--stage', is_flag=True, help='stage the created/updated changelog file with git')
@click.option('-c', '--commit', is_flag=True, help='commit the created/updated changelog file with git, together with other currently staged files')
def changelog(**args):
  """
  Show changelogs or create new entries.
  """

  if (args['version'] or args['reformat']) and args['add']:
    logger.error('unsupported combination of arguments')
    sys.exit(1)

  project.load()
  monorepo = project.monorepo
  package = project.subject if project.subject != monorepo else None

  def _split(s: Optional[str]) -> List[str]:
    return list(filter(bool, map(str.strip, (s or '').split(','))))

  manager = ChangelogManager(
    package.get_changelog_directory() if package
    else monorepo.get_changelog_directory() if monorepo
    else '.changelog')

  if args['add']:

    if not args['for']:
      args['for'] = 'general'

    try:
      type_ = v3.Changelog.Entry.Type[args['add']]
    except KeyError:
      logger.error('invalid changelog type: %r', args['add'])
      sys.exit(1)

    fixes = ['#' + f if f.isdigit() else f for f in _split(args['fixes'])]
    entry = v3.Changelog.Entry(
      type_,
      args['for'],
      args['message'] or '',
      fixes)

    # Allow the user to edit the entry if no description is provided or the
    # -e,--edit option was set.
    if not entry.description or args['edit']:
      serialized = yaml.safe_dump(databind.json.dump(entry, v3.Changelog.Entry, mapper=mapper), sort_keys=False)
      entry = databind.json.load(yaml.safe_load(edit_text(serialized)), v3.Changelog.Entry, mapper=mapper)

    # Validate the entry contents (need a description and at least one type and component).
    if not entry.description or not entry.component:
      logger.error('changelog entries need a component and description')
      sys.exit(1)

    created = not manager.unreleased.exists()
    manager.unreleased.add_entry(entry)
    manager.unreleased.save(create_directory=True)
    message = ('Created' if created else 'Updated') + ' "{}"'.format(manager.unreleased.filename)
    print(colored(message, 'cyan'))

    if args['stage'] or args['commit']:
      _git.add([expect(manager.unreleased.filename)])
    if args['commit']:
      commit_message = entry.description
      if package and monorepo:
        commit_message = '{}({}): '.format(entry.type_.name, package.name) + commit_message
      else:
        commit_message = '{}: '.format(entry.type_.name) + commit_message
      if fixes:
        commit_message += '\n\nfixes ' + ', '.join(fixes)
      _git.commit(commit_message)

    sys.exit(0)

  if args['edit']:
    if not manager.unreleased.exists():
      logger.error('no staged changelog')
      sys.exit(1)
    sys.exit(editor_open(expect(manager.unreleased.filename)))

  changelogs = []
  if args['version'] or not args['all']:
    if args['all']:
      sys.exit('error: incompatible arguments: <version> and -a,--all')
    changelog = manager.version(args['version']) if args['version'] else manager.unreleased
    # Load the changelog for the specified version or the current staged entries.
    if not changelog.exists():
      print('No changelog for {}.'.format(colored(str(args['version'] or 'unreleased'), 'yellow')))
      sys.exit(0)
    changelogs.append(changelog)
  else:
    changelogs = list(manager.all())

  if args['reformat']:
    for changelog in changelogs:
      changelog.save()
    sys.exit(0)

  if args['markdown']:
    changelog_format = 'markdown'
  else:
    changelog_format = 'terminal'

  render_changelogs(sys.stdout, changelog_format, changelogs)



# -*- coding: utf8 -*-
# Copyright (c) 2021 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import logging
import shlex
import subprocess
import sys
from typing import Dict, List

import click
from nr.stream import Stream
from termcolor import colored

from shut.commands import project
from shut.model import dump, PackageModel
from shut.model.requirements import Requirement, VersionSelector, VendoredRequirement
from . import pkg

logger = logging.getLogger(__name__)


@pkg.group()
def requirements():
  """
  Update requirements in the package configuration.
  """


@requirements.command(no_args_is_help=True)
@click.argument('packages', nargs=-1)
@click.option('-v', '--vendor', 'vendored_packages', multiple=True,
  help='Specify a vendored installation source. This can point to a subdirectory of your project '
  'or a Git repository URL in a format that Pip understands (i.e. starting with git+). This is '
  'useful if a dependency is not available on any package index but only exists in a Git '
  'repository (in which case it can be tracked as a submodule and added with this option).')
@click.option('--test', is_flag=True, help='Add the requirement for testing.')
@click.option('--develop/--no-develop', default=True, help='Specifies whether vendored packages '
  'referencing a local directory should be installed in develop mode or not (default: true)')
def add(packages, test, vendored_packages, develop):
  """
  Install packages and save them to the package requirements.

  Note that this will remove any extranous whitespaces, empty lines and comments
  from the package configuration.
  """

  package = project.load_or_exit(expect=PackageModel)
  if not packages and not vendored_packages:
    sys.exit('error: no packages specified')

  reqs = [Requirement.parse(x) for x in packages]
  vendored_reqs = [VendoredRequirement.parse(x, fallback_to_path=True) for x in vendored_packages]
  target = package.test_requirements if test else package.requirements

  # Ensure that the same vendored requirement doesn't already exist.
  duplicate = next((x for x in vendored_reqs if x in target), None)
  if duplicate:
    logger.warning(f'vendored package {colored("%s", "cyan")} already exists', duplicate)

  # Install the requirements with Pip.
  python = shlex.split(os.getenv('PYTHON', 'python'))
  pip = python + ['-m', 'pip']
  command = pip + ['install'] + [r.to_setuptools() for r in reqs]
  command += Stream(x.get_pip_args(package.get_directory(), develop) for x in vendored_reqs).concat()
  command += package.install.get_pip_args()
  res = subprocess.call(command)
  if res != 0:
    sys.exit('error: pip install failed')

  scoped_reqs = [r for r in reqs if r.version != VersionSelector.ANY]
  unscoped_packages = [r.package for r in reqs if r.version == VersionSelector.ANY]

  if unscoped_packages:
    package_versions = get_pip_versions(pip, unscoped_packages)
    scoped_reqs += [Requirement.parse(f'{k} ^{v}') for k, v in package_versions.items()]

  # Update any existing requirements.
  seen = set()
  for req in target.reqs():
    matching_req = next((x for x in scoped_reqs if x.package == req.package), None)
    if matching_req:
      seen.add(req.package)
      req.version = matching_req.version

  # Append the remaining reqs.
  target += [r for r in scoped_reqs if r.package not in seen]
  target += [r for r in vendored_reqs if r not in target]

  dump(package, package.filename)


def get_pip_versions(pip: List[str], package_names: List[str]) -> Dict[str, str]:
  command = pip + ['show'] + package_names
  output = subprocess.check_output(command).decode()
  result = {}
  for chunk in output.split('---'):
    current = {}
    for line in chunk.splitlines():
      key, value = line.partition(':')[::2]
      if key:
        current[key] = value.strip()
    result[current['Name']] = current['Version']
  return result



# -*- coding: utf8 -*-
# Copyright (c) 2021 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import json as _json
import click
from shut.commands import project
from shut.commands.commons.status import get_status, jsonify_status, print_status
from shut.commands.pkg import pkg
from shut.model import PackageModel


@pkg.command(help="""
  Shows whether the package was modified since the last release.
  """ + (print_status.__doc__ or ''))
@click.option('--json', is_flag=True, help='Output as JSON.')
@click.option('--include-config', is_flag=True, help='Include the package config in the JSON output.')
def status(json: bool, include_config: bool) -> None:
  project.load_or_exit(expect=PackageModel)
  status = get_status(project)
  if json:
    result = jsonify_status(project, status, include_config)
    print(_json.dumps(result, indent=2))
  else:
    print_status(status)



# -*- coding: utf8 -*-
# Copyright (c) 2021 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import datetime
import hashlib
import logging
import json
import os
import subprocess as sp
import sys
import traceback
from pathlib import Path
from typing import List, Optional

import click
import typing as t
from databind.core.annotations import typeinfo
from nr.stream import Stream
from termcolor import colored

from shut.commands import shut
from shut.commands.pkg.install import collect_requirement_args, split_extras
from shut.model.package import PackageModel
from shut.test import Runtime, TestRun, TestStatus, Virtualenv
from shut.utils.text import indent_text
from . import pkg, project

log = logging.getLogger(__name__)


class _TestReqsInstalledTracker:
  """
  Internal. Helper class to track if we installed test driver requirements into an environment before.
  We use this as a performance optimization to avoid kicking off `pip install` if we don't need to.
  """

  def __init__(self, package: PackageModel, runtime: Runtime, drivers: t.List[str], isolate: bool) -> None:
    self.package = package
    self.runtime = runtime
    self.drivers = drivers
    self.isolate = isolate

  def get_cache_key(self) -> str:
    return f'{self.package.name}:{self.runtime.get_executable_path()}:' + ','.join(self.drivers) + f':{self.isolate}'

  def get_cache_filename(self) -> Path:
    if self.package.project.monorepo:
      directory = self.package.project.monorepo.get_directory()
    else:
      directory = self.package.get_directory()
    return Path(directory) / 'build' / '.shut' / 'test-reqs-installed-status.json'

  def get_requirements_hash(self, reqs: t.List[str]) -> str:
    reqs_str = ','.join(sorted(reqs)).lower()
    return hashlib.md5(f'{reqs_str}-{self.runtime.get_hash_code()}'.encode()).hexdigest()

  def get_stored_hash(self) -> Optional[str]:
    " Retrieves the currently stored hash. "

    if self.isolate:
      return None

    filename = self.get_cache_filename()
    if not filename.exists():
      return None
    return json.loads(filename.read_text()).get(self.get_cache_key())

  def store_hash(self, hash: str) -> None:
    " Stores a new hash in the cache. "

    if self.isolate:
      return

    filename = self.get_cache_filename()
    data = json.loads(filename.read_text()) if filename.exists() else {}
    data[self.get_cache_key()] = hash
    filename.parent.mkdir(parents=True, exist_ok=True)
    filename.write_text(json.dumps(data))


def test_package(
  package: PackageModel,
  isolate: bool,
  keep_test_env: bool = False,
  capture: bool = True,
  install_test_reqs: t.Optional[bool] = None,
  only: Optional[t.Collection[str]] = None,
  quiet: bool = False,
) -> t.Iterator[t.Tuple[str, TestRun]]:
  drivers = package.get_test_drivers()
  if not drivers:
    raise RuntimeError('package has no test driver configured')

  if only is not None:
    drivers_by_name = {d.NAME: d for d in drivers}
    try:
      drivers = [drivers_by_name[k] for k in only]
    except KeyError as exc:
      raise RuntimeError(f'package has no {colored(str(exc), "red")} test driver configured')

  q = ['-q'] if quiet else []
  venv: Optional[Virtualenv] = None
  if isolate:
    if install_test_reqs is not None and not install_test_reqs:
      log.warning('--no-install is not compatible with --isolate, this will most likely result in an error '
        'when invoking the test drivers.')
    venv = Virtualenv(os.path.join(package.get_directory(), '.venv-test'))
    if venv.exists():
      print('Re-using existing virtual environment at .venv-test ...')
    else:
      print('Creating temporary virtual environment at .venv-test ...')
      venv.create(Runtime.from_env())
    runtime = venv.get_runtime()
    print(f'Installing package "{package.name}"...')
    try:
      orig_cwd = os.getcwd()
      os.chdir(package.get_directory())
      shut(['pkg', '--no-checks', 'install', '--no-develop', '--pip', venv.bin('pip')] + q, standalone_mode=False)
    except SystemExit as exc:
      os.chdir(orig_cwd)
      if exc.code != 0:
        raise
  else:
    runtime = Runtime.from_env()

  test_reqs = Stream(
    [req.to_setuptools() for req in driver.get_test_requirements()]
    for driver in drivers).concat().collect()
  test_reqs += collect_requirement_args(package, develop=False, inter_deps=False, extra={'test'},
    skip_main_requirements=True)
  helper = _TestReqsInstalledTracker(package, runtime, sorted(only or []), isolate)
  reqs_hash = helper.get_requirements_hash(test_reqs)

  if install_test_reqs is None and test_reqs:
    if helper.get_stored_hash() == reqs_hash:
      log.debug('Skipping installation of test driver requirements because it appears we installed them before. This is '
        'a performance optimization that can be skipped using the --install option explicitly.')
    elif not runtime.is_venv():
      log.warning('Skipping installation of test driver requirements because it doesn\'t look like you are in a local '
        'Python environment (environment: %s)', runtime.get_environment())
    else:
      install_test_reqs = True

  if install_test_reqs and test_reqs:
    log.info('Installing test driver requirements %s...', test_reqs)
    sp.check_call(runtime.pip + ['install'] + q + test_reqs)
    helper.store_hash(reqs_hash)

  try:
    for driver in drivers:
      started = datetime.datetime.now()
      try:
        print('[{time}] Running test driver {driver} for package {pkg}'.format(
          driver=colored(driver.NAME, 'cyan'),
          pkg=colored(package.name, 'cyan', attrs=['bold']),
          time=datetime.datetime.now()))
        yield driver.NAME, driver.test_package(package, runtime, capture)
      except Exception:
        yield driver.NAME, TestRun(
          started=started,
          duration=(datetime.datetime.now() - started).total_seconds(),
          status=TestStatus.ERROR,
          environment=runtime.get_environment(),
          tests=[],
          errors=[],
          error=traceback.format_exc()
        )
  finally:
    if venv and not keep_test_env:
      venv.rm()


def test_status_color(status: TestStatus) -> str:
  return {
    TestStatus.PASSED: 'green',
    TestStatus.SKIPPED: 'yellow',
    TestStatus.FAILED: 'red',
    TestStatus.ERROR: 'magenta'
  }[status]


def print_test_run(test_run: TestRun) -> None:
  if test_run.status == TestStatus.ERROR:
    print(f'There was an unexpected error when running the tests.')
    if test_run.error:
      print()
      print(colored(indent_text(test_run.error.strip(), 4), 'red'))
      print()

  sorted_tests = sorted(test_run.tests, key=lambda t: t.name)

  # Print a summary.
  n_passed = sum(1 for t in test_run.tests if t.status == TestStatus.PASSED)
  n_skipped = sum(1 for t in test_run.tests if t.status == TestStatus.SKIPPED)
  status_line = (
    f'Ran {len(test_run.tests)} test(s) in {test_run.duration:.3f}s '
    f'({n_passed} passed, {n_skipped} skipped, {len(test_run.tests) - n_passed - n_skipped} failed, '
    f'{len(test_run.errors)} error(s)). {test_run.status.name}')
  print()
  for test in sorted_tests:
    print(f'  {colored(test.name, test_status_color(test.status), attrs=["bold"])} {test.status.name}')
  if sorted_tests:
    print()

  if (n_passed + n_skipped) < len(test_run.tests):
    # Print error details.
    print('Failed test details:')
    print('====================')
    not_passed_not_skipped = lambda t: t.status not in (TestStatus.PASSED, TestStatus.SKIPPED)
    for i, test in enumerate(filter(not_passed_not_skipped, sorted_tests)):
      print()
      uncolored_line = f'  {test.name} ({test.filename}:{test.lineno})'
      line = f'  {colored(test.name, "red", attrs=["bold"])} ({test.filename}:{test.lineno})'
      print(line)
      print('  ' + '-' * (len(uncolored_line) - 2))
      print()
      if test.crash:
        print(indent_text(test.crash.longrepr, 6))
      if test.stdout:
        print('\n  captured stdout:\n')
        print(indent_text(test.stdout, 6))
    print()

  if test_run.errors:
    header = f'Encountered {len(test_run.errors)} error(s)'
    print(header)
    print('=' * len(header))
    for err in test_run.errors:
      print()
      if err.filename:
        print(f'  {colored(err.filename, "red")}')
        print('  ' + '-' * len(err.filename))
        print()
      else:
        print('  ----')
      print(indent_text(err.longrepr, 6))
    print()

  print(colored(status_line))


@pkg.command()
@click.option('--isolate/--no-isolate', default=False,
  help='Isolate all test runs in virtual environments. This greatly increases the duration '
       'for tests to run as the environment needs to be created and packages installed first, '
       'but it ensures that the unit tests work for a vanilla installation (default: false)')
@click.option('--keep-test-env', is_flag=True,
  help='Do not delete the virtual environment created when testing with --isolate.')
@click.option('--capture/--no-capture', default=True,
  help='Capture the output of the underlying testing framework. If set to false, the output '
       'will be routed to stderr (default: true)')
@click.option('--install/--no-install', default=None,
  help='Install test driver requirements required by the driver. This is enabled by default unless this '
       'command is run not from a virtual environment. Shut performs a minor optimization in that '
       'it skips the installation if it appears to have installed the dependencies before (you can '
       'pass --install explicitly to ensure that the test driver requirements are installed before invoking '
       'the test drivers).')
@click.option('--only', type=split_extras, help='A command-separated list of drivers names to run only.')
@click.option('-q', '--quiet', is_flag=True, help='Quiet Pip install of test requirements.')
def test(
  isolate: bool,
  keep_test_env: bool,
  capture: bool,
  install: t.Optional[bool],
  only: t.Optional[t.Set[str]],
  quiet: bool
) -> None:
  """
  Run the package's unit tests.
  """

  package = project.load_or_exit(expect=PackageModel)
  num_passed = 0
  num_tests = 0

  statuses: List[t.Tuple[str, TestStatus]] = []

  for driver_name, test_run in test_package(package, isolate, keep_test_env, capture, install, only, quiet):
    num_tests += 1
    print_test_run(test_run)
    print()
    statuses.append((driver_name, test_run.status))
    if test_run.status in (TestStatus.PASSED, TestStatus.SKIPPED):
      num_passed += 1

  for driver_name, status in statuses:
    print(f'  {colored(driver_name, test_status_color(status), attrs=["bold"])} {status.name}')

  sys.exit(0 if num_passed == num_tests else 1)




import os
import typing as t


def get_file_in_directory(
  directory: str,
  prefix: str,
  preferred: t.List[str],
  case_sensitive: bool = True,
) -> t.Optional[str]:
  """
  Returns a file in *directory* that is either in the *preferred* list or starts with
  specified *prefix*.
  """

  if not case_sensitive:
    preferred = [x.lower() for x in preferred]

  choices = []
  for name in sorted(os.listdir(directory)):
    if (case_sensitive and name in preferred) or (not case_sensitive and name.lower() in preferred):
      return os.path.join(directory, name)
    if name.startswith(prefix):
      choices.append(name)
  else:
    if choices:
      return choices[0]

  return None



# -*- coding: utf8 -*-
# Copyright (c) 2021 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import abc
import logging
import os
import sys
from collections import Counter
from dataclasses import dataclass
from typing import Iterable, Generic, List, Optional, Type, TypeVar

import click
import nr.fs  # type: ignore
from nr.stream import Stream
from nr.utils.git import Git
from termcolor import colored

from shut.changelog.manager import ChangelogManager
from shut.commands import project
from shut.model import AbstractProjectModel, Project
from shut.model.version import bump_version, parse_version, Version
from shut.renderers import get_version_refs, VersionRef
from shut.utils.io.virtual import VirtualFiles
from shut.utils.text import substitute_ranges

logger = logging.getLogger(__name__)
T_AbstractProjectModel = TypeVar('T_AbstractProjectModel', bound=AbstractProjectModel)


@dataclass
class Args:
  version: Optional[Version]
  major: bool
  minor: bool
  patch: bool
  post: bool
  snapshot: bool
  tag: bool
  push: bool
  dry: bool
  warnings_as_errors: bool
  skip_checks: bool
  skip_update: bool
  force: bool
  allow_lower: bool = False


class VersionBumpData(Generic[T_AbstractProjectModel], metaclass=abc.ABCMeta):

  def __init__(self, args: Args, project: Project, obj: T_AbstractProjectModel) -> None:
    self.args = args
    self.project = project
    self.obj = obj

  def loaded(self) -> None:
    pass

  @abc.abstractmethod
  def get_snapshot_version(self) -> Version:
    pass

  def get_version_refs(self) -> Iterable[VersionRef]:
    yield from get_version_refs(self.obj)

  def bump_to_version(self, target_version: Version) -> Iterable[str]:
    """
    Called to bump to the specified *target_version*. The default implementation uses the
    version refs provided by #get_version_refs() to bump.
    """

    version_refs = list(self.get_version_refs())

    print()
    print(f'bumping {len(version_refs)} version reference(s)')

    for filename, refs in Stream(version_refs).groupby(lambda r: r.filename, lambda it: list(it)):
      with open(filename) as fp:
        content = fp.read()

      if len(refs) == 1:
        ref = refs[0]
        print(f'  {colored(nr.fs.rel(ref.filename), "cyan")}: {ref.value} → {target_version}')
      else:
        print(f'  {colored(nr.fs.rel(ref.filename), "cyan")}:')
        for ref in refs:
          print(f'    {ref.value} → {target_version}')

      content = substitute_ranges(
        content,
        ((ref.start, ref.end, str(target_version)) for ref in refs),
      )

      if not self.args.dry:
        with open(filename, 'w') as fp:
          fp.write(content)

    changed_files = set(x.filename for x in version_refs)

    # Release the staged changelog.
    managers = list(self.get_changelog_managers())
    managers = [m for m in managers if m.unreleased.exists()]
    if managers:
      print()
      print('release staged changelog' + ('s' if len(managers) > 1 else ''))
      for manager in managers:
        assert manager.unreleased.filename
        changed_files.add(manager.unreleased.filename)
        if self.args.dry:
          changelog = manager.version(target_version)
        else:
          changelog = manager.release(target_version)
        assert changelog.filename
        changed_files.add(changelog.filename)
        print(f'  {colored(os.path.relpath(manager.unreleased.filename), "cyan")} → {os.path.relpath(changelog.filename)}')

    return changed_files

  @abc.abstractmethod
  def update(self, new_version: Version) -> VirtualFiles:
    """
    Run the "update" function for the current monorepo or package. A list of the modified files
    must be returned.
    """

  def get_changelog_managers(self) -> Iterable[ChangelogManager]:
    yield ChangelogManager(self.obj.get_changelog_directory())


def make_bump_command(
  data_class: Type[VersionBumpData[AbstractProjectModel]],
  model_type: Type[AbstractProjectModel],
) -> click.Command:

  @click.argument('version', type=parse_version, required=False)
  @click.option('--major', is_flag=True, help='bump the major number')
  @click.option('--minor', is_flag=True, help='bump the minor number')
  @click.option('--patch', is_flag=True, help='bump the patch number')
  @click.option('--post', is_flag=True, help='bump the post-release number')
  @click.option('--snapshot', is_flag=True, help='update the version number by appending the Git commit distance and shasum (note: not compatible with publishing to PyPI)')
  @click.option('-t', '--tag', is_flag=True, help='create a commit and Git tag after bumping the version')
  @click.option('-p', '--push', is_flag=True, help='push the new commit and tag to the Git "origin" remote')
  @click.option('-d', '--dry', is_flag=True, help='do not write changes to disk')
  @click.option('-w', '--warnings-as-errors', is_flag=True, help='treat check warnings as errors')
  @click.option('--skip-checks', is_flag=True, help='skip running checks before bumping')
  @click.option('--skip-update', is_flag=True, help='skip update after bumping')
  @click.option('--force', '-f', is_flag=True, help='force target version (allowing you to "bump" '
    'the same or a lower version). the flag will also result in force adding a Git tag and force '
    'pushing to the remote repository if the respective options as set (--tag and --push).')
  def bump(**kwargs):
    """
    Bump the version number to prepare a new release.
    """

    args = Args(**kwargs)
    data = data_class(args, project, project.load_or_exit(expect=model_type))
    data.loaded()
    do_bump(args, data)

  return bump


def do_bump(args: Args, data: VersionBumpData[AbstractProjectModel]) -> None:
  # Validate arguments.
  if args.push and not args.tag:
    sys.exit('error: --push can only be used with --tag')

  bump_args = 'version major minor patch post snapshot'.split()
  provided_args = {k for k in bump_args if getattr(args, k)}
  if len(provided_args) > 1:
    sys.exit('error: conflicting options: {}'.format(provided_args))
  elif not provided_args:
    print()
    print('figuring bump mode from changelog')
    entries = []
    for manager in data.get_changelog_managers():
      if manager.unreleased.exists():
        entries += manager.unreleased.entries
    if not entries:
      sys.exit('error: no changelog entries found')
    counter = Counter(entry.type_.name for entry in entries)
    bump_mode = max(entry.type_.bump_mode for entry in entries)
    assert hasattr(args, bump_mode.name), bump_mode.name
    setattr(args, bump_mode.name, True)
    print('  {} → {}'.format(
      ', '.join(f'{c} {colored(k, "cyan")}' for k, c in counter.items()),
      colored(bump_mode.name, "blue", attrs=["bold"])))

  version_refs = list(get_version_refs(data.obj))
  if not version_refs:
    sys.exit('error: no version refs found')

  # Ensure the version is the same accross all refs.
  current_version = data.obj.get_version()
  is_inconsistent = any(parse_version(x.value) != current_version for x in version_refs)
  if is_inconsistent:
    if not args.force:
      sys.exit('error: inconsistent versions across files need to be fixed first.')
    logger.warning('inconsistent versions across files were found.')

  # Bump the current version number.
  if args.post:
    assert current_version
    new_version = bump_version(current_version, 'post')
  elif args.patch:
    assert current_version
    new_version = bump_version(current_version, 'patch')
  elif args.minor:
    assert current_version
    new_version = bump_version(current_version, 'minor')
  elif args.major:
    assert current_version
    new_version = bump_version(current_version, 'major')
  elif args.snapshot:
    new_version = data.get_snapshot_version()
    if new_version < current_version:
      # The snapshot version number can be considered lower, so we'll allow it.
      args.allow_lower = True
  else:
    assert args.version, "no new version specified"
    new_version = args.version

  if not new_version.pep440_compliant:
    logger.warning(f'version "{new_version}" is not PEP440 compliant.')

  # The new version cannot be lower than the current one, unless forced.
  if new_version < current_version and not (args.force or args.allow_lower):
    sys.exit(f'version "{new_version}" is lower than current version "{current_version}"')
  if str(new_version) == str(current_version) and not args.force:
    # NOTE(NiklasRosenstein): Comparing as strings to include pre-release and build number.
    logger.warning(f'new version "{new_version}" is equal to current version')
    exit(0)

  # Bump version numbers in files.
  changed_files = list(data.bump_to_version(new_version))

  if not args.skip_update:
    print()
    print('updating files')
    changed_files += data.update(new_version).abspaths('.')

  if args.tag:
    print()
    git = Git()
    files = list(git.porcelain())

    # We require that no files are currently staged.
    if any(f.mode == 'A' for f in files):
      logger.error('cannot tag with non-empty staging area')
      exit(1)

    # TODO: If this step errors (e.g. for example because the unreleased changelog file
    #   was not yet tracked by Git, making Git complain that it doesn't know about the
    #   file and it doesn't exits anymore when we add it down below), it's a bit painful
    #   to recover from it because it's a manual process of reverting the version number
    #   bumps.

    tag_name = data.obj.get_tag(new_version)
    print(f'tagging {tag_name}')

    if not args.dry:
      git.add(changed_files)
      git.commit('({}) bump version to {}'.format(data.obj.get_name(), new_version), allow_empty=True)
      git.tag(tag_name, force=args.force)

    if not args.dry and args.push:
      git.push('origin', git.get_current_branch_name(), tag_name, force=args.force)



# -*- coding: utf8 -*-
# Copyright (c) 2021 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from dataclasses import dataclass, field
import re
from typing import Iterable, Optional

import networkx as nx  # type: ignore

from .abstract import AbstractProjectModel
from .version import Version
from .requirements import VersionSelector
from .release import MonorepoReleaseConfiguration
from .package import PackageModel
from .publish import PublishConfiguration


@dataclass
class InterdependencyRef:
  filename: str
  package_name: str
  version_selector: Optional[VersionSelector]
  version_start: int
  version_end: int


@dataclass
class MonorepoModel(AbstractProjectModel):

  def get_inter_dependencies(self) -> Iterable[InterdependencyRef]:
    """
    Returns an iterable that contains the names of packages in the mono repository to a list
    of their dependencies on other packages in the same repository. Note that it does so
    by regex-matching in the package configuration file rather than reading the deserialized
    package data in order to return start and end index data.
    """

    assert self.project
    for package in self.project.packages:
      yield from self.get_inter_dependencies_for(package)

  def get_inter_dependencies_for(self, package: PackageModel) -> Iterable[InterdependencyRef]:
    """
    Like #get_inter_dependencies() but for a single package.
    """

    assert self.project
    assert package.filename
    regex = re.compile(r'^\s*- +([A-z0-9\.\-_]+) *([^\n:]+)?$', re.M)
    package_names = set(p.name for p in self.project.packages)
    with open(package.filename) as fp:
      content = fp.read()
      for match in regex.finditer(content):
        package_name, version_selector_str = match.groups()
        if package_name not in package_names:
          continue
        version_selector = VersionSelector(version_selector_str) if version_selector_str else None
        yield InterdependencyRef(package.filename, package_name, version_selector, match.start(2), match.end(2))

  def get_inter_dependencies_graph(self) -> nx.DiGraph:
    """
    Create a directed graph from the inter dependencies of packages in the mono repo.
    """

    assert self.project
    graph = nx.DiGraph()
    for package in self.project.packages:
      graph.add_node(package.name)
    for package in self.project.packages:
      for ref in self.get_inter_dependencies_for(package):
        graph.add_edge(ref.package_name, package.name)

    return graph

  # AbstractProjectModel Overrides

  publish: PublishConfiguration = field(default_factory=PublishConfiguration)
  release: MonorepoReleaseConfiguration = field(default_factory=MonorepoReleaseConfiguration)

  def get_name(self) -> str:
    return self.name

  def get_version(self) -> Optional[Version]:
    return self.version
